# pylint: disable=invalid-name

from struct import pack, unpack

DEFAULT_COMPRESSION_LEVEL = 20


def rotate_left(a, b, bits=32):
    a &= (1 << bits) - 1
    return ((((a) << (b)) | ((a) >> (bits - (b))))) & ((1 << bits) - 1)


def quarter_round(w, a, b, c, d):
    w[a] += w[b]
    w[a] &= 0xFFFFFFFF
    w[d] ^= w[a]
    w[d] = rotate_left(w[d], 16)
    w[c] += w[d]
    w[c] &= 0xFFFFFFFF
    w[b] ^= w[c]
    w[b] = rotate_left(w[b], 12)
    w[a] += w[b]
    w[a] &= 0xFFFFFFFF
    w[d] ^= w[a]
    w[d] = rotate_left(w[d], 8)
    w[c] += w[d]
    w[c] &= 0xFFFFFFFF
    w[b] ^= w[c]
    w[b] = rotate_left(w[b], 7)


def round_pair(w):
    quarter_round(w, 0, 4, 8, 12)
    quarter_round(w, 1, 5, 9, 13)
    quarter_round(w, 2, 6, 10, 14)
    quarter_round(w, 3, 7, 11, 15)
    quarter_round(w, 0, 5, 10, 15)
    quarter_round(w, 1, 6, 11, 12)
    quarter_round(w, 2, 7, 8, 13)
    quarter_round(w, 3, 4, 9, 14)


def two_rounds(s):
    w = s_to_w(s)
    round_pair(w)
    round_pair(w)
    return w_to_s(w)


def quad_rounds(s):
    w = s_to_w(s)
    round_pair(w)
    round_pair(w)
    round_pair(w)
    round_pair(w)
    return w_to_s(w)


def sxor(x, y):
    return bytes(a ^ b for a, b in zip(x, y))


def s_to_w(s):
    return list(unpack(b"<" + (b"I" * (len(s) // 4)), s))


def w_to_s(w):
    return pack(b"<" + (b"I" * len(w)), *w)


def s_to_q(s):
    return list(unpack(b"<" + (b"Q" * (len(s) // 8)), s))


def q_to_s(w):
    return pack(b"<" + (b"Q" * len(w)), *w)


def add_qwords(h0, h1):
    return q_to_s(
        [(a + b) & 0xFFFFFFFFFFFFFFFF for a, b in zip(s_to_q(h0), s_to_q(h1))]
    )


def mix_in(h, s):
    def mix_partial(h, partial):
        assert len(partial) <= 0x40
        b = bytearray(h)
        for i, c in enumerate(partial[::-1]):
            b[i] ^= ord("%c" % c)
        return quad_rounds(bytes(b))

    while s != b"":
        h = mix_partial(h, s[:0x40])
        s = s[0x40:]

    return h


def hash_filepath_v1(filepath):
    # Generate initial hash from the string
    h0 = mix_in(b"\x00" * 0x40, filepath)

    # Advance h0 by four round pairs to get h1
    h1 = quad_rounds(h0)
    # Add the two together, and advance by four round pairs.
    key = quad_rounds(add_qwords(h0, h1))

    return keyed_hashing(filepath, key)


def hash_filepath_v2(filepath, key):
    # Generate initial hash from the string
    h = two_rounds(pack(b"<QQQQQQQQ", key, len(filepath), 0, 0, 0, 0, 0, 0))
    h = mix_in(h, filepath)

    # Add the two together, and advance by four round pairs.
    tmp = add_qwords(h, quad_rounds(h))
    tmp = s_to_q(tmp)
    key = quad_rounds(q_to_s([tmp[0] ^ len(filepath)] + tmp[1:]))

    return keyed_hashing(filepath, key)


def hash_filepath(filepath, key=None, version="v2"):
    if version == "v1":
        return hash_filepath_v1(filepath)

    if version == "v2":
        return hash_filepath_v2(filepath, key)

    raise ValueError("Invalid version provided.")


def keyed_hashing(filepath, key):
    # Do keyed hashing
    # NOTE: This appears to be an implementation mistake on the Spelunky 2 dev's part
    # They generate a quad_round advanced version of (nonce'd key), but then they
    # xor with the untweaked key instead of the tweaked key...

    h = b""
    for i in range(0, len(filepath), 0x40):
        partial = filepath[i : i + 0x40]
        h += sxor(partial, key[: len(partial)][::-1])
    return h


class Key:
    def __init__(self, key=0, mask=2 ** 64 - 1):
        self.key = key
        self.mask = mask

    def update(self, asset_len):
        v3 = (
            0x9E6C63D0676A9A99
            * (
                self.key
                ^ asset_len
                ^ rotate_left(self.key ^ asset_len, 17, 64)
                ^ rotate_left(self.key ^ asset_len, 64 - 25, 64)
            )
            & self.mask
        )
        v4 = 0x9E6D62D06F6A9A9B * (v3 ^ ((v3 ^ (v3 >> 28)) >> 23)) & self.mask
        self.key ^= v4 ^ ((v4 ^ (v4 >> 28)) >> 23)


def mix_in_filepath(filepath, h):
    # Mix the filename in to tweak the key
    for i in range(0, len(filepath), 0x40):
        partial = filepath[i : i + 0x40]
        h = quad_rounds(sxor(h[: len(partial)], partial[::-1]) + h[len(partial) :])
    return h


def chacha_rest(data, key):
    # NOTE: This appears to be an implementation mistake on the Spelunky 2 dev's part
    # They generate a quad_round advanced version of (nonce'd key), but then they
    # xor with the untweaked key instead of the tweaked key...
    out = b""
    if len(data) >= 0x40:
        blocks = len(data) // 0x40
        out += sxor(data, key[::-1] * blocks)
        data = data[blocks * 0x40 :]
    if len(data) > 0:
        out += sxor(data, key[: len(data)][::-1])

    return out


def chacha_v1(filepath, data):
    # Untweaked key begins as half-advanced "0xBABE"
    h = two_rounds(pack(b"<QQQQQQQQ", 0xBABE, 0, 0, 0, 0, 0, 0, 0))

    h = mix_in_filepath(filepath, h)

    # Add the tweaked key and its advancement, then advance by four round pairs.
    key = quad_rounds(add_qwords(h, quad_rounds(h)))

    return chacha_rest(data, key)


def chacha_v2(filepath, data, key):
    # Untweaked key begins as half-advanced `key`
    h = two_rounds(pack(b"<QQQQQQQQ", key, len(filepath), 0, 0, 0, 0, 0, 0))

    h = mix_in_filepath(filepath, h)

    # Add the tweaked key and its advancement, then advance by four round pairs.
    tmp = add_qwords(h, quad_rounds(h))
    tmp = s_to_q(tmp)
    key = quad_rounds(q_to_s([tmp[0] ^ key + len(data)] + tmp[1:]))

    return chacha_rest(data, key)


def chacha(filepath, data, key=None, version="v2"):
    if version == "v1":
        return chacha_v1(filepath, data)

    if version == "v2":
        return chacha_v2(filepath, data, key)

    raise ValueError("Invalid version provided.")



from .full_sheets import *


def get_all_sprite_mergers(base_path: str):
    return [
        CharacterBlackSpriteMerger(base_path),
        CharacterLimeSpriteMerger(base_path),
        CharacterMagentaSpriteMerger(base_path),
        CharacterOliveSpriteMerger(base_path),
        CharacterOrangeSpriteMerger(base_path),
        CharacterPinkSpriteMerger(base_path),
        CharacterRedSpriteMerger(base_path),
        CharacterVioletSpriteMerger(base_path),
        CharacterWhiteSpriteMerger(base_path),
        CharacterYellowSpriteMerger(base_path),
        CharacterBlueSpriteMerger(base_path),
        CharacterCeruleanSpriteMerger(base_path),
        CharacterCinnabarSpriteMerger(base_path),
        CharacterCyanSpriteMerger(base_path),
        CharacterGoldSpriteMerger(base_path),
        CharacterGraySpriteMerger(base_path),
        CharacterGreenSpriteMerger(base_path),
        CharacterIrisSpriteMerger(base_path),
        CharacterKhakiSpriteMerger(base_path),
        CharacterLemonSpriteMerger(base_path),
        CharacterEggChildSpriteMerger(base_path),
        CharacterHiredHandSpriteMerger(base_path),
        TurkeySpriteMerger(base_path),
        RockdogSpriteMerger(base_path),
        AxolotlSpriteMerger(base_path),
        QilinSpriteMerger(base_path),
        MontySpriteMerger(base_path),
        PercySpriteMerger(base_path),
        PoochiSpriteMerger(base_path),
    ]



import logging
from functools import wraps

from modlunky2.assets.patcher import Patcher

logger = logging.getLogger("modlunky2")


def is_patched(exe_filename):
    with exe_filename.open("rb") as exe:
        return Patcher(exe).is_checksum_patched()


def log_exception(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            logger.exception("Unexpected failure")

    return wrapper



from pathlib import Path

from ..base_classes.base_sprite_loader import BaseSpriteLoader
from ..util import chunks_from_animation


class Mounts(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/mounts.png")
    _chunk_size = 128
    _chunk_map = {
        "turkey": (0, 0, 1, 1),
        **chunks_from_animation("turkey_row_0_", (0, 0, 1, 1), 14),
        **chunks_from_animation("turkey_row_1_", (0, 1, 1, 2), 16),
        **chunks_from_animation("turkey_row_2_", (0, 2, 1, 3), 16),
        **chunks_from_animation("turkey_row_3_", (0, 3, 1, 4), 14),
        "rockdog": (0, 4, 1, 5),
        **chunks_from_animation("rockdog_row_0_", (0, 4, 1, 5), 14),
        **chunks_from_animation("rockdog_row_1_", (0, 5, 1, 6), 12),
        **chunks_from_animation("rockdog_row_2_", (0, 6, 1, 7), 15),
        **chunks_from_animation("rockdog_row_3_", (0, 7, 1, 8), 14),
        "axolotl": (0, 8, 1, 9),
        **chunks_from_animation("axolotl_row_0_", (0, 8, 1, 9), 14),
        **chunks_from_animation("axolotl_row_1_", (0, 9, 1, 10), 12),
        **chunks_from_animation("axolotl_row_2_", (0, 10, 1, 11), 15),
        **chunks_from_animation("axolotl_row_3_", (0, 11, 1, 12), 14),
        "qilin": (0, 12, 1, 13),
        **chunks_from_animation("qilin_row_0_", (0, 12, 1, 13), 14),
        **chunks_from_animation("qilin_row_1_", (0, 13, 1, 14), 12),
        **chunks_from_animation("qilin_row_2_", (0, 14, 1, 15), 15),
        **chunks_from_animation("qilin_row_3_", (0, 15, 1, 16), 14),
    }



from .big import Big1, Big2, Big3, Big4, Big5, Big6
from .ghost import Ghost
from .basic import Basic1, Basic2, Basic3
from .mounts import Mounts
from .pets import Pets

__all__ = [
    "Big1",
    "Big2",
    "Big3",
    "Big4",
    "Big5",
    "Big6",
    "Ghost",
    "Basic1",
    "Basic2",
    "Basic3",
    "Mounts",
    "Pets",
]



import hashlib

def md5sum_path(path, chunk_size=8192):
    """ Streaming md5 digest from a path."""

    with path.open("rb") as file_:
        md5sum = hashlib.md5()
        chunk = file_.read(chunk_size)
        while chunk:
            md5sum.update(chunk)
            chunk = file_.read(chunk_size)
        return md5sum.hexdigest().encode()



from dataclasses import dataclass
from collections import OrderedDict
from typing import ClassVar, Optional, TextIO, Tuple, Union

from .utils import DirectivePrefixes, parse_chance_values, split_comment, to_line

VALID_MONSTER_CHANCES = set(
    [
        "bat",
        "bee",
        "cat",
        "caveman",
        "cobra",
        "crabman",
        "critteranchovy",
        "critterbutterfly",
        "crittercrab",
        "critterdrone",
        "critterdungbeetle",
        "critterfirefly",
        "critterfish",
        "critterlocust",
        "critterpenguin",
        "critterslime",
        "crittersnail",
        "crocman",
        "female_jiangshi",
        "firebug",
        "firefrog",
        "fish",
        "frog",
        "giantfly",
        "giantspider",
        "hangspider",
        "hermitcrab",
        "hornedlizard",
        "imp",
        "jiangshi",
        "landmine",
        "lavamander",
        "leprechaun",
        "mantrap",
        "mole",
        "monkey",
        "mosquito",
        "necromancer",
        "octopus",
        "olmite",
        "robot",
        "snake",
        "sorceress",
        "spider",
        "springtrap",
        "tadpole",
        "tikiman",
        "ufo",
        "vampire",
        "witchdoctor",
        "yeti",
    ]
)

NAME_PADDING = max(map(len, VALID_MONSTER_CHANCES)) + 4


class MonsterChances:
    def __init__(self):
        self._inner = OrderedDict()
        self.comment = None

    def get(self, name):
        MonsterChance.validate_name(name)
        return self._inner.get(name)

    def set_obj(self, obj: "MonsterChance"):
        obj.clean()
        obj.validate()
        self._inner[obj.name] = obj

    def write(self, handle: TextIO):
        if self.comment:
            handle.write(f"{self.comment}\n")
        for obj in self._inner.values():
            handle.write(obj.to_line())
        handle.write("\n")


@dataclass
class MonsterChance:
    prefix: ClassVar[str] = DirectivePrefixes.MONSTER_CHANCE.value
    name: str
    value: Union[int, Tuple[int, int, int, int]]
    comment: Optional[str]

    @classmethod
    def parse(cls, line: str) -> "MonsterChance":
        rest, comment = split_comment(line)
        directive, value = rest.split(None, 1)
        name = directive[2:]

        if not name:
            raise ValueError("Directive missing name.")

        obj = cls(name, value, comment)
        obj.clean()
        obj.validate()

        return obj

    @staticmethod
    def validate_name(name: str):
        if name not in VALID_MONSTER_CHANCES:
            raise ValueError(f"Invalid name provided for MonsterChance: {name!r}")

    def validate_value(self):
        if isinstance(self.value, list) and len(self.value) != 4:
            raise ValueError(
                f"Expected 4 values but got {len(self.value)} for {self.name}"
            )
        elif not isinstance(self.value, (list, int)):
            raise ValueError(f"Got unexpected type for {self.name}: {self.value!r}")

    def validate(self):
        self.validate_name(self.name)
        self.validate_value()

    def clean_value(self):
        if not isinstance(self.value, str):
            return

        value = parse_chance_values(self.value)

        if len(value) not in (1, 4):
            raise ValueError(
                f"MonsterChance {self.name!r} has {len(value)} values when 1 or 4 is expected: {value!r}"
            )

        if len(value) == 1:
            value = value[0]

        self.value = value

    def clean(self):
        self.clean_value()

    def value_to_str(self) -> str:
        value = self.value
        if isinstance(value, int):
            value = [value]
        return ", ".join(map(str, value))

    def to_line(self) -> str:
        return to_line(
            self.prefix, self.name, NAME_PADDING, self.value_to_str(), 28, self.comment
        )

    def write(self, handle: TextIO):
        handle.write(self.to_line())



from pathlib import Path

from ..base_classes import BaseSpriteLoader


class Basic1(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbasic01.png")
    _chunk_size = 128
    _chunk_map = {
        "snake": (0, 0, 1, 1),
        "bat": (6, 0, 7, 1),
        "skeleton": (0, 2, 1, 3),
        "leprechaun": (0, 4, 1, 5),
        "spider": (0, 3, 1, 4),
        "shopkeep": (0, 6, 1, 7),
        "ufo": (0, 8, 1, 9),
        "alien": (0, 9, 1, 10),
        "cobra": (0, 10, 1, 11),
        "scorpion": (0, 12, 1, 13),
        "golden_monkey": (14, 9, 15, 10),
        "bee": (9, 12, 10, 13),
        "magmar": (4, 14, 5, 15),
    }


class Basic2(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbasic02.png")
    _chunk_size = 128
    _chunk_map = {
        "vampire": (6, 1, 7, 2),
        "vlad": (6, 3, 7, 4),
        "caveman": (0, 5, 1, 6),
        "caveman_asleep": (8, 6, 9, 7),
        "bodyguard": (0, 8, 1, 9),
        "oldhunter": (0, 11, 1, 12),
        "tun": (0, 14, 1, 15),
    }


class Basic3(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbasic03.png")
    _chunk_size = 128
    _chunk_map = {
        "beg": (0, 0, 1, 1),
        "thief": (0, 1, 1, 2),
        "sister": (0, 5, 1, 6),
        "sister1": (0, 11, 1, 12),
        "sister2": (0, 8, 1, 9),
        "sister3": (0, 5, 1, 6),
        "yang": (0, 14, 1, 15),
    }



r"""

Library for patching out checksum checking for assets.

This is the code that validates the checksum and calls exit() when it doesn't match:
                  /------------------------------------------\
cmp rax,rcx  jz -/   xor ecx,ecx  call cs:exit        int 3   \-> mov rcx,[rbp+17h]
48 3B C1     74 09   33 C9        FF ?? ?? ?? ?? ??   CC          48 8B 4D 17
48 3B C1     74 09   90 90        90 90 90 90 90 90   90
                     [               nop               ]

We overwrite the exit() call with NOPs
"""

import logging

CHECKSUM_PATCH_START = b"\x48\x3B\xC1\x74\x09\x33\xC9\xFF"
CHECKSUM_PATCH_END = 0xCC
CHECKSUM_PATCH_REPLACE = b"\x48\x3B\xC1\x74\x09" + b"\x90" * 9

RELEASE_AOB_PRODUCTION = b"\x00\x50\x72\x6F\x64\x75\x63\x74\x69\x6F\x6E\x00"
RELEASE_AOB_REPLACE = b"\x00\x4D\x6F\x64\x6C\x75\x6E\x6B\x79\x32\x00\x00"


logger = logging.getLogger("modlunky2")


class Patcher:
    def __init__(self, exe_handle):
        self.exe_handle = exe_handle

    def find(self, needle, offset=0, bsize=4096):
        if bsize < len(needle):
            raise ValueError(
                "The buffer size must be larger than the string being searched for."
            )

        self.exe_handle.seek(offset)
        overlap = len(needle) - 1

        while True:
            buffer = self.exe_handle.read(bsize)
            if not buffer:
                return -1

            pos = buffer.find(needle)
            if pos >= 0:
                return self.exe_handle.tell() - len(buffer) + pos

            if len(buffer) <= overlap:
                return -1

            self.exe_handle.seek(self.exe_handle.tell() - overlap)

    def is_checksum_patched(self) -> bool:
        """ Returns true of the binary has already been patched."""
        self.exe_handle.seek(0)
        return self.find(CHECKSUM_PATCH_START) == -1

    def patch_checksum(self):
        self.exe_handle.seek(0)

        logger.info("Patching asset checksum check")
        index = self.find(CHECKSUM_PATCH_START)
        if index == -1:
            logger.warning("Didn't find instructions to patch. Is game unmodified?")
            return False

        self.exe_handle.seek(index)
        ops = self.exe_handle.read(14)

        if ops[-1] != CHECKSUM_PATCH_END:
            logger.warning(
                "Checksum check has unexpected form, this script has "
                "to be updated for the current game version."
            )
            logger.warning(
                "(Expected 0x{:02x}, found 0x{:02x})".format(
                    CHECKSUM_PATCH_END, ops[-1]
                )
            )
            return False

        logger.info("Found check at 0x{:08x}, replacing with NOPs".format(index))
        self.exe_handle.seek(index)
        self.exe_handle.write(CHECKSUM_PATCH_REPLACE)

    def patch_release(self):
        self.exe_handle.seek(0)
        index = self.find(RELEASE_AOB_PRODUCTION)
        if index == -1:
            logger.warning(
                "Didn't find production string in release. Is this a vanilla binary being patched?"
            )
            return False

        self.exe_handle.seek(index)
        self.exe_handle.write(RELEASE_AOB_REPLACE)



from .items import ItemSheet
from .journal_stickers import StickerSheet
from .journal_items import JournalItemSheet
from .journal_people import JournalPeopleSheet
from .journal_mons import JournalMonsterSheet
from .journal_mons_big import JournalBigMonsterSheet
from .journal_place import JournalPlaceSheet
from .journal_traps import JournalTrapSheet
from .character import *
from .monsters.mounts import Mounts
from .monsters.pets import Pets


def get_all_sprite_loaders(base_path: str):
    return [
        ItemSheet(base_path),
        StickerSheet(base_path),
        JournalItemSheet(base_path),
        JournalPeopleSheet(base_path),
        JournalMonsterSheet(base_path),
        JournalBigMonsterSheet(base_path),
        JournalPlaceSheet(base_path),
        JournalTrapSheet(base_path),
        JournalTrapSheet(base_path),
        CharacterBlackSheet(base_path),
        CharacterLimeSheet(base_path),
        CharacterMagentaSheet(base_path),
        CharacterOliveSheet(base_path),
        CharacterOrangeSheet(base_path),
        CharacterPinkSheet(base_path),
        CharacterRedSheet(base_path),
        CharacterVioletSheet(base_path),
        CharacterWhiteSheet(base_path),
        CharacterYellowSheet(base_path),
        CharacterBlueSheet(base_path),
        CharacterCeruleanSheet(base_path),
        CharacterCinnabarSheet(base_path),
        CharacterCyanSheet(base_path),
        CharacterEggChildSheet(base_path),
        CharacterGoldSheet(base_path),
        CharacterGraySheet(base_path),
        CharacterGreenSheet(base_path),
        CharacterHiredHandSheet(base_path),
        CharacterIrisSheet(base_path),
        CharacterKhakiSheet(base_path),
        CharacterLemonSheet(base_path),
        Mounts(base_path),
        Pets(base_path),
    ]



from typing import Tuple, List
from enum import Enum


def split_comment(line: str) -> Tuple[str, str]:
    rest, _, comment = line.partition("//")
    comment = comment.strip().strip("/")
    return (rest.strip(), comment)


def parse_chance_values(values: str) -> List[int]:
    return [int(value.strip()) for value in values.split(",")]


def to_line(prefix, name, name_padding, value, value_padding, comment) -> str:
    line = f"{prefix}{name:<{name_padding}}"
    if comment:
        line = f"{line}{value:<{value_padding}} // {comment}"
    else:
        line = f"{line}{value}"
    return f"{line}\n"


class Peekable:
    def __init__(self, iterable):
        self._buffer = []
        self._iterator = iter(iterable)

    def __iter__(self):
        return self

    def __next__(self):
        if not self._buffer:
            return next(self._iterator)
        return self._buffer.pop(0)

    def peek(self):
        try:
            val = next(self._iterator)
        except StopIteration:
            return None
        self._buffer.append(val)
        return val

    def advance(self):
        try:
            val = next(self)
        except StopIteration:
            return None
        return val


class DirectivePrefixes(Enum):
    LEVEL_SETTING = r"\-"
    TILE_CODE = r"\?"
    LEVEL_CHANCE = r"\%"
    MONSTER_CHANCE = r"\+"
    TEMPLATE = r"\."
    TEMPLATE_SETTING = r"\!"



import argparse
import logging
from os.path import exists
from pathlib import Path

from appdirs import user_config_dir, user_data_dir
from setuptools.command.bdist_egg import INSTALL_DIRECTORY_ATTRS

from .constants import APP_DIR
from .ui import ModlunkyUI

APP_AUTHOR = "spelunky.fyi"
APP_NAME = "modlunky2"


def make_dirs(*dirs):
    for dir_ in dirs:
        if not dir_.exists():
            dir_.mkdir(parents=True, exist_ok=True)


class Config:
    def __init__(self, config_path):
        self.config_path = config_path

        self.install_dir = None

    @classmethod
    def from_path(cls, config_path):
        obj = cls(config_path=config_path)

        if config_path.exists():
            with config_path.open("r") as config_file:
                config_data = json.load(config_file)
        else:
            config_data = {}

        obj.install_dir = config_data.get("install-dir")

        return obj

    def to_dict(self):
        return {
            "install-dir": self.install_dir,
        }

    def _get_tmp_path(self):
        return self.config_path.with_suffix(f"{self.config_path.suffix}.tmp")

    def save(self):
        tmp_path = self._get_tmp_path()




def main():
    parser = argparse.ArgumentParser(description="Tool for modding Spelunky 2.")
    parser.add_argument(
        "--install-dir",
        default=APP_DIR,
        help="Path to Spelunky 2 installation. (Default: %(default)s",
    )
    parser.add_argument(
        "--beta", default=False, action="store_true", help="Display beta features."
    )
    args = parser.parse_args()

    log_format = "%(asctime)s: %(message)s"
    logging.basicConfig(format=log_format, level=logging.INFO, datefmt="%H:%M:%S")

    install_dir = Path(args.install_dir)

    config_dir = Path(user_config_dir(APP_NAME, APP_AUTHOR))
    data_dir = Path(user_data_dir(APP_NAME, APP_AUTHOR))

    make_dirs(config_dir, data_dir)

    native_ui = ModlunkyUI(install_dir, args.beta)
    native_ui.mainloop()



from pathlib import Path

from ..base_classes.base_sprite_loader import BaseSpriteLoader


class Ghost(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monsters_ghost.png")
    _chunk_size = 128
    _chunk_map = {"ghist_shopkeeper": (7, 10, 8, 11), "ghost": (0, 0, 2, 2)}
